#! /usr/bin/env python3
# 
# WPScan output parser
# 
# Authors: Florian Roth, Tristan Landès
#
# DISCLAIMER - USE AT YOUR OWN RISK.
# 
# Some infos are intentionally ignored when parsing Json to have shorter output.
# You can use --format cli to show all informations with Infos, Warnings and Alerts
# 
# Exemple stdin usage:
#   $ wpscan --url https://exemple.com --format json | python3 ./parser.py
#
# With param --input :
#   $ python3 ./parser.py --input wpscan.log
#
# Or you can import this package into your application and call `parse_results` method.
#   from wpwatcher.parser import parse_results
#   (messages, warnings, alerts) = parse_results(wpscan_output_string)

"""
All the WPScan fields for the JSON output in the views/json folders at:

https://github.com/wpscanteam/CMSScanner/tree/master/app/views/json
https://github.com/wpscanteam/wpscan/tree/master/app/views/json

Here are some other inspirational ressources found about parsing wpscan json

Generates a nice table output (Rust code) 
    https://github.com/lukaspustina/wpscan-analyze
    Parser code: 
        https://github.com/lukaspustina/wpscan-analyze/blob/master/src/analyze.rs
Python parser (do not parse for vulnerable theme or outdated warnings) 
    https://github.com/aaronweaver/AppSecPipeline/blob/master/tools/wpscan/parser.py
Vulcan wpscan (Go) 
    https://github.com/adevinta/vulcan-checks/blob/master/cmd/vulcan-wpscan/wpscan.go
    Great job listing all the fields, is the list complete ?
Dradis ruby json Parser 
    https://github.com/dradis/dradis-wpscan/blob/master/lib/dradis/plugins/wpscan/importer.rb : 
    No warnings neither but probably the clearest code

Ressource PArsing CLI output:
    List of all icons: https://github.com/wpscanteam/CMSScanner/blob/master/app/formatters/cli.rb
"""

import json
import re
import sys
import argparse

def parse_results(wpscan_output, false_positives=[]):
    # Init scan messages
    ( messages, warnings, alerts ) = ([],[],[])
    is_json=False
    try:
        data=json.loads(wpscan_output)
        is_json=True
    except: pass
    if is_json:
        (messages, warnings, alerts)=parse_json(data)
    else: 
        (messages, warnings, alerts)=parse_cli(wpscan_output)
    #Process false positives
    for alert in alerts:
        if is_false_positive(alert, false_positives):
            alerts.remove(alert)
            messages.append("[False positive]\n"+alert)

    for warn in warnings:
        if is_false_positive(warn, false_positives):
            warnings.remove(warn)
            messages.append("[False positive]\n"+warn)
    
    return (( messages, warnings, alerts ))

def parse_cli(wpscan_output):
    if "[+]" in wpscan_output:
        # Init scan messages
        ( messages, warnings, alerts ) = ([],[],[])
        # Init messages toogles
        warning_on = False
        alert_on = False
        # Test if cli_with_colors
        cli_with_colors= ( "32m[+]" in wpscan_output )
        message="" 
        # Every blank ("") line will be considered as a message separator
        for line in wpscan_output.splitlines():

            # Parse all output lines and build infos, warnings and alerts
            line=line.strip()

            # Empty content lines are ignored
            # Parse the line and Toogle Warning/Alert
            if line!="" and line!="|":   

                # Toogle Warning/Alert if specific match in any line of the message
                if cli_with_colors==False:

                    # Method 1 : No color. Warnings string are hard coded here -------------
                    # Remove colorization
                    # line = re.sub(r'(\x1b|\[[0-9][0-9]?m)','',line)
                    if "[!]" in line and any([m in line for m in [   
                        "The version is out of date",
                        "No WPVulnDB API Token given",
                        "You can get a free API token"]]) :
                        warning_on = True
                    elif "[!]" in line :
                        alert_on = True

                    # Method 2 : Cli with colors parsing ------------------------------------
                else: 
                    if "33m[!]" in line: warning_on=True
                    if "31m[!]" in line: alert_on = True
                    # Remove colorization anyway after parsing
                    line = re.sub(r'(\x1b|\[[0-9][0-9]?m)','',line)

                # Both method with color and no color apply supplementary proccessing 
                # Warning for insecure Wordpress
                if 'Insecure' in line: 
                    warning_on = True
                # Lower voice of Vulnerabilities found but not plugin version
                if 'The version could not be determined' in line and alert_on:
                        alert_on = False  
                        warning_on = True 

                # When line message has been read and parsed
                # Append line to message. Handle the begin of the message case
                message+= line if message=="" else '\n'+line 

            # Message separator just a white line.
            elif line=="":
                # Only if the message if not empty. 
                # End of the message
                if message.strip() != "":
                    # Append messages to list of infos, warns and alerts
                    if alert_on: alerts.append(message)
                    elif warning_on: warnings.append(message)
                    else: messages.append(message)
                    message=""
                    # Reset Toogle Warning/Alert
                    alert_on = False  
                    warning_on = False

        # Catching last message after loop
        if alert_on: alerts.append(message)
        elif warning_on: warnings.append(message)
        else: messages.append(message)
        return (( messages, warnings, alerts ))
    else: 
        raise Exception("The file does not seem to be a WPScan CLI log.")

def parse_vulnerabilities_and_warnings_only(data):
    alerts=[]
    warnings=[]
    for section in data:
        node=data[section]
        if section=='version':
            warnings.extend(parse_warning_wordpress(node))
        if section=='main_theme':
            warnings.extend(parse_warning_theme_or_plugin(node))
        if any ([section==c for c in ['main_theme','version']]):
            try: alerts.extend(parse_findings(node['vulnerabilities']))
            except: pass
        if any([section==c for c in ['themes', 'plugins', 'timthumbs']]):
            for slug in node:
                try: alerts.extend(parse_findings(node[slug]['vulnerabilities']))
                except: pass
                try: warnings.extend(parse_warning_theme_or_plugin(node[slug]))
                except: pass
    return ((warnings, alerts))

def parse_json(data):
    # Do a sanity check to confirm the data is ok
    if data and 'target_url' in data and data['target_url']:
        warnings, alerts=parse_vulnerabilities_and_warnings_only(data)
        infos, w1, a1=parse_other_infos(data)
        warnings.extend(w1)
        alerts.extend(a1)
        return (( infos, warnings, alerts ))
    else: 
        raise Exception("No data in wpscan Json output (None) or no 'target_url' field present in the provided Json data. The scan might have failed, data: \n"+str(data))
    
def parse_other_infos(data):
    messages=[]
    warnings=[]
    alerts=[]
    messages.append("Target URL: {}\nIP: {}\nEffective URL: {}".format(
        data['target_url'],
        data["target_ip"],
        data["effective_url"]))
    
    if "banner" in data:
        messages.append("Scanned with WPScan version {}".format(data['banner']['version']))

    if "last_db_update" in data:
        messages.append("Last WPScan database update: {}".format(data['last_db_update']))

    # Parse know vulnerabilities
    # Parse vulnerability data and make more human readable.
    # NOTE: You need an API token for the WPVulnDB vulnerability data.

    if "interesting_findings" in data:
        if data["interesting_findings"] is not None and len(data["interesting_findings"])>0:
            # Parse informations
            messages.extend(parse_findings(data["interesting_findings"]) )

    if "users" in data:
        if data["users"] is not None and len(data["users"])>0:
            users = data["users"]
            for name in users:
                # Parse users
                messages.append( 'WordPress user found: %s'%name )
    
    if "config_backups" in data:
        if data["config_backups"] is not None or len(data["config_backups"])>0:
            for url in data['config_backups']:
                alerts.append("WordPress Configuration Backup Found\nURL: %s"%str(url) )

    if "db_exports" in data :
        if data['db_exports'] is not None and len(data['db_exports'])>0:
            alerts.extend(parse_vulnerability_or_finding(data['db_exports'] ))

    if "password_attack" in data :
        if data['password_attack'] is not None and len(data['password_attack'])>0:
            for passwd in data['password_attack']:
                alerts.append("WordPres Weak User Password Found:\n%s"%str(passwd) )

    if "medias" in data :
        if data['medias'] is not None and len(data['medias'])>0:
            warnings.extend(parse_vulnerability_or_finding(data['medias']))

    if "vuln_api" in data :
        if "error" in data['vuln_api']:
            warnings.append(data['vuln_api']["error"])

    if "not_fully_configured" in data and data['not_fully_configured']!=None :
        alerts.append(data['not_fully_configured'])

    return (( messages, warnings, alerts ))

    # Older code not ised anymore

    # if "main_theme" in data:
    #     if data["main_theme"]==None or len(data["main_theme"])==0:
    #         messages.append("WPScan did not find any main theme information")
    #     else:
    #         # Parse theme warnings
    #         warnings.extend(parse_warning_theme_or_plugin('main theme',data['main_theme']) )

    #         if "vulnerabilities" in data["main_theme"]:
    #             # Parse Vulnerable themes
    #             alerts.extend(parse_findings('Vulnerable theme',data["main_theme"]["vulnerabilities"]) )

    #         if "version" in data["main_theme"] and data["main_theme"]["version"] != None and "vulnerabilities" in data["main_theme"]["version"] :
    #             # Parse vulnerable theme version
    #             alerts.extend(parse_findings('Vulnerable theme',data["main_theme"]["version"]["vulnerabilities"]) )
    
    # if "version" in data:
    #     if data["version"]==None or len(data["version"])==0:
    #         messages.append("WPScan did not find any WordPress version")
    #     else:
    #         # Parse WordPress version
    #         messages.append(parse_version_info(data['version']))
    #         # Parse outdated WordPress version
    #         warnings.extend(parse_warning_wordpress(data['version']) )
    #         # Parse vulnerable WordPress version
    #         alerts.extend(parse_findings('Vulnerable wordpress',data["version"]["vulnerabilities"]) )
    
    # if "themes" in data:
    #     if data["themes"]==None or len(data["themes"])==0:
    #         messages.append("WPScan did not find any theme information")
    #     else:
    #         for theme in data["themes"]:
    #             # Parse secondary theme warnings
    #             warnings.extend(parse_warning_theme_or_plugin('theme',theme) )
    #             # Parse secondary Vulnerable themes
    #             alerts.extend(parse_findings('Vulnerable theme', data["themes"][theme]["vulnerabilities"]) )
    
    # if "plugins" in data:
    #     if data["plugins"]==None or len(data["plugins"])==0:
    #         messages.append("WPScan did not find any WordPress plugins")
    #     else:
    #         plugins = data["plugins"]
    #         for plugin in plugins:
    #             # Parse vulnerable plugins
    #             alerts.extend(parse_findings('Vulnerable pulgin',plugins[plugin]["vulnerabilities"]) )
    #             # Parse outdated plugins
    #             warnings.extend(parse_warning_theme_or_plugin('plugin',plugins[plugin]) )

def parse_vulnerability_or_finding(finding):
    # Finding can be a vulnerability or other
    findingData = ""
    refData = ""
    title=""
    # title = "%s:"%(finding_type) if finding_type else ""

    if type(finding) is dict:
        # For interesting findings
        # if "type" in finding:
        #     title += "[%s]" % finding["type"]
        if "to_s" in finding:
            title += "%s" % finding["to_s"]
        # For vulnerabilities
        if "title" in finding:
            title += "%s" % finding["title"]

        findingData += "%s" % title

        if "fixed_in" in finding:
            findingData += "\nFixed In: %s" % finding["fixed_in"]

        if "url" in finding:
            findingData += "\nURL: %s" % finding["url"]

        # if "found_by" in finding:
        #     findingData += "\nFound by: %s" % finding["found_by"]

        if "confidence" in finding:
            findingData += "\nConfidence: %s" % finding["confidence"]

        if "interesting_entries" in finding:
            if len(finding["interesting_entries"]) > 0:
                findingData += "\nInteresting Entries: %s" % (", ".join(finding["interesting_entries"]))

        # if "comfirmed_by" in finding:
        #     if len(finding["confirmed_by"]) > 0:
        #         findingData += "\nConfirmed By:\n"
        #         findingData+="\n- ".join(finding["confirmed_by"])

        if "references" in finding and len(finding["references"]) > 0:
            refData += "\nReferences:"
            for ref in finding["references"]:
                if ref =='cve':
                    for cve in finding["references"][ref]: refData+="\n- CVE: http://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-%s"%(cve)
                elif ref == 'wpvulndb': 
                    for wpvulndb in finding["references"][ref]: refData+="\n- wpvulndb: https://wpvulndb.com/vulnerabilities/%s"%(wpvulndb)
                else:
                    refData += "\n- %s: %s" % (ref, ", ".join(finding["references"][ref]) )

    else: raise TypeError("Must be a dict, method parse_a_finding() for data {}".format(finding)) 
    return ("%s %s" % (findingData, refData) )

# Wrapper to parse findings can take list or dict type
def parse_findings(findings):
    summary = []
    if type(findings) is list:
        for finding in findings:
            summary.append(parse_vulnerability_or_finding(finding))
    elif type(findings) is dict:
        for finding in findings:
            summary.append(parse_vulnerability_or_finding(findings[finding]))
    else:
        raise TypeError("Must be a list or dict, method parse_findings() for data: {}".format(findings)) 
    return(summary)

# def parse_version_info(version):
#     headerInfo = ""

#     if "number" in version:
#         headerInfo += "Running WordPress version: %s\n" % version["number"]

#     if "interesting_entries" in version:
#             if len(version["interesting_entries"]) > 0:
#                 headerInfo += "\nInteresting Entries: %s" % (", ".join(version["interesting_entries"]))

#     return headerInfo

def parse_warning_wordpress(finding):
    summary=[]
    if not finding: return summary
    warn=False
    findingData=""
    if 'status' in finding and finding['status']=="insecure":
        findingData+="Insecure WordPress version %s identified (released on %s)"%(finding['number'], finding['release_date'])
        warn=True

    # if "found_by" in finding:
    #         findingData += "\nFound by: %s" % finding["found_by"]

    if "confidence" in finding:
        findingData += "\nConfidence: %s" % finding["confidence"]

    # if "interesting_entries" in finding:
    #         if len(finding["interesting_entries"]) > 0:
    #             findingData += "\nInteresting Entries: %s" % (", ".join(finding["interesting_entries"]))

    if warn: summary.append(findingData)
    return(summary)

def parse_warning_theme_or_plugin(finding):
    summary=[]
    if not finding: return summary
    warn=False
    findingData=""

    if 'slug' in finding:
        findingData+="%s" % finding['slug']

    if 'outdated' in finding and finding['outdated']==True:
        findingData+="\nThe version is out of date, the latest version is %s" % (finding["latest_version"])
        warn=True

    if "directory_listing" in finding and finding['directory_listing']:
        findingData+="\nDirectory listing is enabled"
        warn=True

    if "error_log_url" in finding and finding['error_log_url']:
        findingData+="\nAn error log file has been found: %s" % (finding["error_log_url"])
        warn=True

    if "location" in finding:
        findingData += "\nLocation: %s" % finding["location"]

    # if "found_by" in finding:
    #     findingData += "\nFound by: %s" % finding["found_by"]

    if "confidence" in finding:
        findingData += "\nConfidence: %s" % finding["confidence"]

    if "interesting_entries" in finding:
        if len(finding["interesting_entries"]) > 0:
            findingData += "\nInteresting Entries: %s" % (", ".join(finding["interesting_entries"]))

    if warn: summary.append(findingData)
    return(summary)

# False Positive Detection
def is_false_positive(string, false_positives):
    for fp_string in false_positives:
        if fp_string in string:
            return True
    return False

def parse_args():
    parser = argparse.ArgumentParser(description='WPscan output parser')
    parser.add_argument('--input', '-i', metavar='path', help="WPScan Json or CLI output")
    args = parser.parse_args()
    return args

if __name__ == '__main__':
    # Init scan messages
    ( messages, warnings, alerts ) = ([],[],[])
    args=parse_args()
    if args.input:
        # Parse file
        with open(args.input) as wpout:
            (infos, warnings, alerts)=parse_results( wpout.read() , [] )
    else:
        # Parse stdin
        lines = sys.stdin.readlines()
        for i in range(len(lines)):
            lines[i] = lines[i].replace('\n','')
        (infos, warnings, alerts)=parse_results( '\n'.join(lines) , [] )
        
    # Building message
    if (warnings or alerts) :message = "Issues have been detected by WPScan.\n"
    else: message = "WPScan report\n"
    if alerts:
        message += "\n\n\tAlerts\n\n"
        message += "\n\n".join(alerts)
    if warnings:
        message += "\n\n\tWarnings\n\n"
        message += "\n\n".join(warnings)
    if infos:
        message += "\n\n\tInformations\n\n"
        message += "\n\n".join(infos)
    print(message)



"""
Wordpress Watcher
Automating WPscan to scan and report vulnerable Wordpress sites

DISCLAIMER - USE AT YOUR OWN RISK.
"""

# Main program, parse the args, read config and launch scans
import time
import argparse
import shlex
import json
import sys

from wpwatcher import VERSION, AUTHORS, GIT_URL, log
from wpwatcher.utils import init_log, parse_timedelta, results_summary
from wpwatcher.config import WPWatcherConfig
from wpwatcher.core import WPWatcher

class WPWatcherCLI():
    def __init__(self):
        args=self.parse_args()
        init_log(args.verbose, args.quiet)
        # If template conf , print and exit
        if args.template_conf:
            print(WPWatcherConfig.TEMPLATE_FILE)
            exit(0)
        log.info("WPWatcher -  Automating WPscan to scan and report vulnerable Wordpress sites")
        # If version, print and exit
        if args.version:
            log.info("Version:\t\t%s"%VERSION)
            log.info("Authors:\t\t%s"""%AUTHORS)
            exit(0)
        # Init WPWatcher obhect and dump reports
        if args.wprs!=False:
            if args.wprs==None :
                f=WPWatcher(WPWatcherConfig().build_config()[0]).find_wp_reports_file()
            else:
                f=args.wprs
            log.info("Reports: %s"%(f))
            with open(f) as r:
                results=json.load(r)
            print(results_summary(results))
            exit(0)
            
        # Read config
        configuration=self.build_config_cli(args)
        # Create main object
        wpwatcher=WPWatcher(configuration)
        # If daemon lopping
        if wpwatcher.conf['daemon']: 
            log.info("Daemon mode selected, looping for ever...")
            results=None # Keep databse in memory
            while True:
                # Run scans for ever
                exit_code,results=wpwatcher.run_scans_and_notify()
                log.info("Daemon sleeping %s and scanning again..."%wpwatcher.conf['daemon_loop_sleep'])
                time.sleep(wpwatcher.conf['daemon_loop_sleep'].total_seconds())
                wpwatcher=WPWatcher(self.build_config_cli(args))
                wpwatcher.wp_reports=results
        # Run scans and quit
        else:
            exit_code,results=wpwatcher.run_scans_and_notify()
            exit(exit_code)

    # Arguments can overwrite config file values
    @staticmethod
    def parse_args():
        parser = argparse.ArgumentParser(description="""WordPress Watcher is a Python wrapper for WPScan that manages scans on multiple sites and reports by email.
    Some config arguments can be passed to the command.
    It will overwrite previous values from config file(s).
    Check %s for more informations."""%(GIT_URL), formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('--conf', '-c', metavar='File path', help="""The script * must read a configuration file to set mail server settings, WPScan path and arguments *.     
    If no config file is found, mail server settings, WPScan path and arguments and other config values will have default values.  
    Setup mail server settings and turn on `send_email_report` in the config file if you want to receive reports.  
    You can specify multiple files `--conf File path [File path ...]`. Will overwrites the keys with each successive file.
    All options keys can be missing from config file.
    If not specified with `--conf` parameter, will try to load config from file `~/.wpwatcher/wpwatcher.conf`, `~/wpwatcher.conf` and `./wpwatcher.conf`.
    All options can be missing from config file.\n\n""", nargs='+', default=None)
        parser.add_argument('--template_conf', '--tmpconf', help="""Print a template config file.
    Use `wpwatcher --template_conf > ~/wpwatcher.conf && vim ~/wpwatcher.conf` to create (or overwrite) and edit the new default config file.""", action='store_true')
        parser.add_argument('--version', '-V', help="Print WPWatcher version", action='store_true')
        parser.add_argument('--wprs', metavar="Path to json file", help="wp_reports database summary generator", nargs='?', default=False)

        # Declare arguments that will overwrite config options
        parser.add_argument('--wp_sites', '--url', metavar="URL", help="Configure wp_sites", nargs='+', default=None)
        parser.add_argument('--wp_sites_list', '--urls', metavar="File path", help="Configure wp_sites from a list of URLs", default=None)
        parser.add_argument('--email_to', '--em', metavar="Email", help="Configure email_to", nargs='+', default=None)
        parser.add_argument('--send_email_report', '--send', help="Configure send_email_report=Yes", action='store_true')
        parser.add_argument('--send_infos', '--infos', help="Configure send_infos=Yes", action='store_true')
        parser.add_argument('--send_errors', '--errors', help="Configure send_errors=Yes", action='store_true')
        parser.add_argument('--attach_wpscan_output', '--attach', help="Configure attach_wpscan_output=Yes", action='store_true')
        parser.add_argument('--fail_fast', '--ff', help="Configure fail_fast=Yes", action='store_true')
        parser.add_argument('--api_limit_wait', '--wait', help="Configure api_limit_wait=Yes", action='store_true')
        parser.add_argument('--daemon',  help="Configure daemon=Yes", action='store_true')
        parser.add_argument('--daemon_loop_sleep','--loop', metavar='Time string', help="Configure daemon_loop_sleeps")
        parser.add_argument('--wp_reports', '--reports', metavar="File path", help="Configure wp_reports", default=None)
        parser.add_argument('--resend_emails_after','--resend', metavar="Time string", help="Configure resend_emails_after")
        parser.add_argument('--asynch_workers','--workers', metavar="Number of asynchronous workers", help="Configure asynch_workers", type=int)
        parser.add_argument('--log_file','--log', metavar="Logfile path", help="Configure log_file")
        parser.add_argument('--follow_redirect','--follow',  help="Configure follow_redirect=Yes", action='store_true')
        parser.add_argument('--wpscan_output_folder','--wpout', metavar="WPScan results folder", help="Configure wpscan_output_folder")
        parser.add_argument('--wpscan_args','--wpargs', metavar='WPScan arguments as string', help='Configure wpscan_args')
        parser.add_argument('--false_positive_strings','--fpstr', metavar='False positive strings', help='Configure false_positive_strings', nargs='+', default=None)
        parser.add_argument('--verbose', '-v', help="Configure verbose=Yes", action='store_true')
        parser.add_argument('--quiet', '-q', help="Configure quiet=Yes", action='store_true')
        args = parser.parse_args()
        return(args)

    # Assemble the config dict from args and from file
    @staticmethod
    def build_config_cli(args):
        args=vars(args) if hasattr(args, '__dict__') and not type(args)==dict else args
        # Configuration variables
        conf_files=args['conf'] if 'conf' in args else None
        # Init config dict: read config files
        configuration, files =WPWatcherConfig(files=conf_files).build_config()
        if files: log.info("Load config file(s) : %s"%files)
        conf_args={}
        # Sorting out only args that matches config options and that are not None or False
        for k in args: 
            if k in WPWatcherConfig.DEFAULT_CONFIG.keys() and args[k]:
                conf_args.update({k:args[k]})  
        # Append or init list of urls from file if any
        if 'wp_sites_list' in args and args['wp_sites_list'] :
            with open(args['wp_sites_list'], 'r') as urlsfile:
                sites=[ site.replace('\n','') for site in urlsfile.readlines() ]
                conf_args['wp_sites']= sites if 'wp_sites' not in conf_args else conf_args['wp_sites']+sites
        # Adjust special case of urls that are list of dict
        if 'wp_sites' in conf_args:
            conf_args['wp_sites']=[ {"url":site} for site in conf_args['wp_sites'] ]
        # Adjust special case of resend_emails_after
        if 'resend_emails_after' in conf_args:
            conf_args['resend_emails_after']=parse_timedelta(conf_args['resend_emails_after'])
        # Adjust special case of daemon_loop_sleep
        if 'daemon_loop_sleep' in conf_args:
            conf_args['daemon_loop_sleep']=parse_timedelta(conf_args['daemon_loop_sleep'])
        # Adjust special case of wpscan_args
        if 'wpscan_args' in conf_args:
            conf_args['wpscan_args']=shlex.split(conf_args['wpscan_args'])
        # if vars(args)['resend']: conf_args['resend_email_after']=timedelta(seconds=0)
        # Overwrite with conf dict biult from CLI Args
        if conf_args: configuration.update(conf_args)
        return configuration

def main(): 
    WPWatcherCLI()

if __name__ == '__main__':
    WPWatcherCLI()


"""
Wordpress Watcher
Automating WPscan to scan and report vulnerable Wordpress sites

DISCLAIMER - USE AT YOUR OWN RISK.
"""
import shlex
import os 
import traceback
import subprocess
from subprocess import CalledProcessError
from wpwatcher import log
from wpwatcher.utils import safe_log_wpscan_args, oneline
# WPScan helper class -----------
class WPScanWrapper():

    def __init__(self, path):
        self.path=path
        # List of current WPScan processes
        self.processes=[]

    # Helper method: actually wraps wpscan
    def wpscan(self, *args):
        (exit_code, output)=(0,"")
        # WPScan arguments
        cmd=shlex.split(self.path) + list(args)
        # Log wpscan command without api token
        log.debug("Running WPScan command: %s" % ' '.join(safe_log_wpscan_args(cmd)) )
        # Run wpscan -------------------------------------------------------------------
        try:
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=open(os.devnull,'w') )
            # Append process to current process list and launch
            self.processes.append(process)
            wpscan_output, _  = process.communicate()
            self.processes.remove(process)
            try: wpscan_output=wpscan_output.decode("utf-8")
            except UnicodeDecodeError: wpscan_output=wpscan_output.decode("latin1")
            # Error when wpscan failed, except exit code 5: means the target has at least one vulnerability.
            #   See https://github.com/wpscanteam/CMSScanner/blob/master/lib/cms_scanner/exit_code.rb
            if process.returncode not in [0,5]:
                # Handle error
                err_string="WPScan failed with exit code %s %s" % (str(process.returncode), ". WPScan output: %s"%wpscan_output if wpscan_output else '')
                log.error(oneline(err_string))
            else :
                # WPScan comamnd success
                log.debug("WPScan raw output:\n"+wpscan_output)
            (exit_code, output)=(process.returncode, wpscan_output)
        except (CalledProcessError) as err:
            # Handle error --------------------------------------------------
            wpscan_output=str(err.output)
            err_string="WPScan failed with exit code %s. WPScan output: \n%s\nError:\n%s" % (str(process.returncode), wpscan_output, traceback.format_exc())
            log.error(oneline(err_string))
            (exit_code, output)=(err.returncode, wpscan_output)
        except FileNotFoundError as err:
            err_string="Could not find wpscan executable. \nError:\n%s" % (traceback.format_exc())
            log.error(oneline(err_string))
            (exit_code, output)=(-1, "")
        return((exit_code, output))
    
    # Check if WPScan is installed
    def is_wpscan_installed(self):
        exit_code, _ = self.wpscan("--version")
        if exit_code!=0: return False
        else: return True

    # Update WPScan database
    def update_wpscan(self):
        log.info("Updating WPScan")
        exit_code, _ = self.wpscan("--update")
        if exit_code!=0: 
            log.error("Error updating WPScan")
            exit(-1)



"""
Wordpress Watcher
Automating WPscan to scan and report vulnerable Wordpress sites

DISCLAIMER - USE AT YOUR OWN RISK.
"""
import logging
import re
import os
import sys
import socket
import copy
from datetime import timedelta

from wpwatcher import VERSION, log

# Few static helper methods -------------------

# Setup stdout logger
def init_log(verbose=False, quiet=False, logfile=None, nostd=False):
    format_string='%(asctime)s - %(levelname)s - %(message)s'
    format_string_cli='%(levelname)s - %(message)s'
    if verbose : verb_level=logging.DEBUG
    elif quiet : verb_level=logging.ERROR
    else : verb_level=logging.INFO
    # Add stdout: configurable
    log.setLevel(verb_level)
    std = logging.StreamHandler(sys.stdout)
    std.setLevel(verb_level)
    std.setFormatter(logging.Formatter(format_string_cli))
    log.handlers=[]
    if not nostd: log.addHandler(std)
    else: log.addHandler(logging.StreamHandler(open(os.devnull,'w')))
    if logfile :
        fh = logging.FileHandler(logfile)
        fh.setLevel(verb_level)
        fh.setFormatter(logging.Formatter(format_string))
        log.addHandler(fh)
    if verbose and quiet :
        log.info("Verbose and quiet values are both set to True. By default, verbose value has priority.")
    return (log)

# Replace --api-token param with *** for safe logging
def safe_log_wpscan_args(wpscan_args):
    logged_cmd=copy.deepcopy(wpscan_args)
    if "--api-token" in logged_cmd :
        logged_cmd[logged_cmd.index("--api-token")+1]="***"
    return logged_cmd

# Helper method that transform multiline string to one line for grepable output
def oneline(string):
    return( " ".join(line.strip() for line in string.splitlines()) )

# Return the given string converted to a string that can be used for a clean filename
def get_valid_filename(s):
    s = str(s).strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)

# Build the email report string
def build_message(wp_report):
    
    message="WordPress security scan report for site: %s\n" % (wp_report['site'])
    message+="Scan datetime: %s\n" % (wp_report['datetime'])
    
    if wp_report['errors'] : message += "\nAn error occurred."
    elif wp_report['alerts'] : message += "\nVulnerabilities have been detected by WPScan."
    elif wp_report['warnings']: message += "\nIssues have been detected by WPScan."
    if wp_report['fixed']: message += "\nSome issues have been fixed since last scan."
    
    if wp_report['errors']:
        message += "\n\n\tErrors\n\t------\n\n"
        message += "\n\n".join(wp_report['errors'])
    if wp_report['alerts']:
        message += "\n\n\tAlerts\n\t------\n\n"
        message += "\n\n".join(wp_report['alerts'])
    if wp_report['fixed']:
        message += "\n\n\tFixed\n\t-----\n\n"
        message += "\n\n".join(wp_report['fixed'])
    if wp_report['warnings']:
        message += "\n\n\tWarnings\n\t--------\n\n"
        message += "\n\n".join(wp_report['warnings'])
    if wp_report['infos']:
        message += "\n\n\tInformations\n\t------------\n\n"
        message += "\n\n".join(wp_report['infos'])
    
    message += "\n\n--"
    message += "\nWPWatcher -  Automating WPscan to scan and report vulnerable Wordpress sites"
    message += "\nServer: %s - Version: %s\n"%(socket.gethostname(),VERSION)
    return message

def print_progress_bar(count,total):
    size=0.3 #size of progress bar
    percent = int(float(count)/float(total)*100)
    log.info( "Progress - [{}{}] {}% - {} / {}".format('='*int(int(percent)*size), ' '*int((100-int(percent))*size), percent, count, total) )

def results_summary(results):
    string='Results summary\n'
    header = ("Site", "Status", "Last email", "Issues", "Problematic component(s)")
    sites_w=20
    # Determine the longest width for site column
    for r in results:
        sites_w=len(r['site'])+4 if r and len(r['site'])>sites_w else sites_w
    frow="{:<%d} {:<8} {:<20} {:<8} {}"%sites_w
    string+=frow.format(*header)
    for row in results:
        pb_components=[]
        for m in row['alerts']+row['warnings']+row['errors']:
            pb_components.append(m.splitlines()[0])
        string+='\n'
        string+=frow.format(row['site'], 
            row['status'],
            str(row['last_email']),
            len(row['alerts']+row['warnings']+row['errors']),
            ', '.join(pb_components) )
    return string

def parse_timedelta(time_str):
    """
    Parse a time string e.g. (2h13m) into a timedelta object.
    """
    regex = re.compile(r'^((?P<days>[\.\d]+?)d)?((?P<hours>[\.\d]+?)h)?((?P<minutes>[\.\d]+?)m)?((?P<seconds>[\.\d]+?)s)?$')
    parts = regex.match(time_str)
    assert parts is not None, "Could not parse any time information from '{}'.  Examples of valid strings: '8h', '2d8h5m20s', '2m4s'".format(time_str)
    time_params = {name: float(param) for name, param in parts.groupdict().items() if param}
    return timedelta(**time_params)



#! /usr/bin/env python3
from setuptools import setup
import pathlib
import re
import sys
import subprocess
# Helper method that will parse wpwatcher.py to extract config setup values
def parse_setup(key):
    part={}
    for line in WPWATCHER.splitlines():
        if key in line:
            exec(line, part)
            break
    return(part[key])
# The directory containing this file
HERE = pathlib.Path(__file__).parent
# Read and store wpwatcher.py file
WPWATCHER = (HERE / "wpwatcher" / "__init__.py").read_text()
# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name                =   'wpwatcher',
    description         =   "WordPress Watcher is a Python wrapper for WPScan that manages scans on multiple sites and reports by email.",
    url                 =   parse_setup('GIT_URL'),
    maintainer          =   parse_setup('AUTHORS'),
    version             =   parse_setup('VERSION'),
    packages            =   ['wpwatcher',], 
    entry_points        =   {'console_scripts': ['wpwatcher = wpwatcher.cli:main'],},
    classifiers         =   ["Programming Language :: Python :: 3"],
    license             =   'Apache License 2.0',
    long_description    =   README,
    long_description_content_type   =   "text/markdown"
)


""""
Wordpress Watcher
Automating WPscan to scan and report vulnerable Wordpress sites

DISCLAIMER - USE AT YOUR OWN RISK.
"""
import copy
import shutil
import traceback
import os
import json
import threading
import time
import io
import concurrent.futures
import unicodedata
import smtplib
import re
import subprocess
import signal
from urllib.parse import urljoin, urlparse, urlunparse
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from datetime import datetime, timedelta

from wpwatcher import log
from wpwatcher.parser import parse_results
from wpwatcher.scan import WPScanWrapper
from wpwatcher.utils import init_log, safe_log_wpscan_args, build_message, get_valid_filename, print_progress_bar, oneline, results_summary

# Wait when API limit reached
API_WAIT_SLEEP=timedelta(hours=24)
# Writing into the database file is thread safe
wp_report_lock = threading.Lock()

# WPWatcher class ---------------------------------------------------------------------
class WPWatcher():

    # WPWatcher must use a configuration dict
    def __init__(self, conf):
        # Copy config dict as is. Copy not to edit initial dict
        self.conf=copy.deepcopy(conf)
        # (Re)init logger with config
        init_log(verbose=self.conf['verbose'],
            quiet=self.conf['quiet'],
            logfile=self.conf['log_file'])
        # Dump config
        log.info("WordPress sites and configuration:{}".format(self.dump_config()))
        # Create wpscan link
        self.wpscan=WPScanWrapper(path=self.conf['wpscan_path'])
        # Check if WPScan exists
        if not self.wpscan.is_wpscan_installed():
            log.error("There is an issue with your WPScan installation or WPScan not installed. Make sure wpscan in you PATH or configure full path to executable in config files. If you're using RVM, the path should point to the WPScan wapper like /usr/local/rvm/gems/ruby-2.6.0/wrappers/wpscan . Fix wpscan on your system. See https://wpscan.org for installation steps.")
            exit(-1)
        # Update wpscan database
        self.wpscan.update_wpscan()
        # Try delete temp files.
        if os.path.isdir('/tmp/wpscan'):
            try: 
                shutil.rmtree('/tmp/wpscan')
                log.info("Deleted temp WPScan files in /tmp/wpscan/")
            except (FileNotFoundError, OSError, Exception) : 
                log.info("Could not delete temp WPScan files in /tmp/wpscan/. Error:\n%s"%(traceback.format_exc()))
        # Read DB
        self.wp_reports=self.build_wp_reports()
        # Try if local Json databse is accessible
        try: self.update_and_write_wp_reports(self.wp_reports)
        except:
            log.error("Could not write wp_reports database: {}. Use '--reports null' to ignore local Json database".format(self.conf['wp_reports']))
            raise
        
        # Init wpscan output folder
        if self.conf['wpscan_output_folder'] : 
            os.makedirs(self.conf['wpscan_output_folder'], exist_ok=True)

        # Executor, will be created when calling run_scans_and_notify
        self.executor=None
        self.futures=[]
        self.scanned_sites=[]
        self.interrupting=False # Toogle if aborting so other errors doesnt get triggerred if using --ff
        # register the signals to be caught ^C and kill will trigger interrupt()
        signal.signal(signal.SIGINT, self.interrupt)
        signal.signal(signal.SIGTERM, self.interrupt)

    def dump_config(self):
        bump_conf=copy.deepcopy(self.conf)
        string=''
        for k in bump_conf:
            v=bump_conf[k]
            if k == 'wpscan_args':
                v=safe_log_wpscan_args(v)
            if k == 'smtp_pass' and bump_conf[k] != "" :
                v = '***'
            if isinstance(v, (list, dict)):
                v=json.dumps(v)
            else: v=str(v)
            string+=("\n{:<25}\t=\t{}".format(k,v))
        return(string)

    def find_wp_reports_file(self, create=False):
        wp_reports=None
        if 'APPDATA' in os.environ: 
            if self.conf['daemon']: p=os.path.join(os.environ['APPDATA'],'.wpwatcher/wp_reports.daemon.json')
            else: p=os.path.join(os.environ['APPDATA'],'.wpwatcher/wp_reports.json')
            if os.path.isfile(p) or create: wp_reports=p
        elif 'HOME' in os.environ: 
            if self.conf['daemon']: p=os.path.join(os.environ['HOME'],'.wpwatcher/wp_reports.daemon.json')
            else: p=os.path.join(os.environ['HOME'],'.wpwatcher/wp_reports.json')
            if os.path.isfile(p) or create: wp_reports=p
        elif 'XDG_CONFIG_HOME' in os.environ: 
            if self.conf['daemon']: p=os.path.join(os.environ['XDG_CONFIG_HOME'],'.wpwatcher/wp_reports.daemon.json')
            else: p=os.path.join(os.environ['XDG_CONFIG_HOME'],'.wpwatcher/wp_reports.json')
            if os.path.isfile(p) or create: wp_reports=p
        # if os.path.isfile('./wp_reports.json'): 
        #     wp_reports='./wp_reports.json'
        if create:
            os.makedirs(os.path.join(os.environ['HOME'],'.wpwatcher'), exist_ok=True)
            if not os.path.isfile(wp_reports):
                with open(wp_reports,'w') as reportsfile:
                    json.dump([],reportsfile)
                    log.info("Init new wp_reports database: %s"%(wp_reports))
        return(wp_reports)
    # Read wp_reports database
    def build_wp_reports(self):
        wp_reports=[]
        if self.conf['wp_reports']!='null':
            if not self.conf['wp_reports']:
                self.conf['wp_reports']=self.find_wp_reports_file(create=True)
            if self.conf['wp_reports']:
                if os.path.isfile(self.conf['wp_reports']):
                    try:
                        with open(self.conf['wp_reports'], 'r') as reportsfile:
                            wp_reports=json.load(reportsfile)
                        log.info("Load wp_reports database: %s"%self.conf['wp_reports'])
                    except Exception:
                        log.error("Could not read wp_reports database: {}. Use '--reports null' to ignore local Json database".format(self.conf['wp_reports']))
                        raise
                else:
                    log.info("The database file %s do not exist. It will be created."%(self.conf['wp_reports']))
        return wp_reports

    def update_and_write_wp_reports(self, new_wp_report_list=[]):
        # Update the sites that have been scanned, keep others
        # Keep same report order add append new sites at the bottom
        for newr in new_wp_report_list:
            new=True
            for r in self.wp_reports:
                if r['site']==newr['site']:
                    self.wp_reports[self.wp_reports.index(r)]=newr
                    new=False
                    break
            if new: 
                self.wp_reports.append(newr)
        # Write to file if not null
        if self.conf['wp_reports']!='null':
            # Write method should be thread safe
            while wp_report_lock.locked():
                time.sleep(0.01)
                continue
            wp_report_lock.acquire()
            with open(self.conf['wp_reports'],'w') as reportsfile:
                json.dump(self.wp_reports, reportsfile, indent=4)
                wp_report_lock.release()
    
    # Send email report with status and timestamp
    def send_report(self, wp_site, wp_report):
        # To
        if len(self.conf['email_errors_to'])>0 and wp_report['status']=='ERROR':
            to_email = ','.join( self.conf['email_errors_to'] )
        else: 
            to_email = ','.join( wp_site['email_to'] + self.conf['email_to'] )
        
        if to_email != "":
           
            # Building message
            message = MIMEMultipart("html")
            message['Subject'] = 'WPWatcher %s report - %s - %s' % (  wp_report['status'], wp_site['url'], wp_report['datetime'])
            message['From'] = self.conf['from_email']
            message['To'] = to_email

            # Email body
            body=build_message(wp_report)
            message.attach(MIMEText(body))
            
            # Attachment log if attach_wpscan_output
            if self.conf['attach_wpscan_output']:
                # Remove color
                wp_report['wpscan_output'] = re.sub(r'(\x1b|\[[0-9][0-9]?m)','', str(wp_report['wpscan_output']))
                # Read the WPSCan output
                attachment=io.BytesIO(wp_report['wpscan_output'].encode())
                part = MIMEBase("application", "octet-stream")
                part.set_payload(attachment.read())
                # Encode file in ASCII characters to send by email    
                encoders.encode_base64(part)
                # Sanitize WPScan report filename 
                wpscan_report_filename=get_valid_filename('WPScan_results_%s_%s' % (wp_site['url'], wp_report['datetime']))
                # Add header as key/value pair to attachment part
                part.add_header(
                    "Content-Disposition",
                    "attachment; filename=%s.txt"%(wpscan_report_filename),
                )
                # Attach the report
                message.attach(part)

            # Connecting and sending
            log.info("Sending %s to %s" % (message['Subject'], to_email))

            # SMTP Connection
            s = smtplib.SMTP(self.conf['smtp_server'])
            s.ehlo()
            # SSL
            if self.conf['smtp_ssl']:
                s.starttls()
            # SMTP Auth
            if self.conf['smtp_auth']:
                s.login(self.conf['smtp_user'], self.conf['smtp_pass'])
            # Send Email
            s.sendmail(self.conf['from_email'], to_email, message.as_string())
            s.quit()
            # Store report time
            wp_report['last_email']=datetime.now().strftime('%Y-%m-%dT%H-%M-%S')
            # Discard fixed items
            wp_report['fixed']=[]
        else:
            log.info("Not sending WPWatcher %s email report because no email are configured for site %s"%(wp_report['status'], wp_site['url']))
    
    def update_report(self, wp_report, last_wp_report):
        # Fill out fixed issues and last_email datetime
        # Save already fixed issues but not reported yet
        wp_report['fixed']=last_wp_report['fixed']
        # Figure out fixed issues : compare firt line of alerts and warnings and see if they are still present
        for last_alert in last_wp_report['alerts']:
            if last_alert.splitlines()[0] not in [a.splitlines()[0] for a in wp_report['alerts']]:
                wp_report['fixed'].append('Alert regarding component "%s" has been fixed since last report.\nLast report sent the %s.\nFix detected the %s'%(last_alert.splitlines()[0], 
                    last_wp_report['last_email'], wp_report['datetime']))
        if self.conf['send_warnings']:
            for last_warn in last_wp_report['warnings']:
                if last_warn.splitlines()[0] not in [a.splitlines()[0] for a in wp_report['warnings']]:
                    wp_report['fixed'].append('Warning regarding component "%s" has been fixed since last report.\nLast report sent the %s.\nFix detected the %s'%(last_warn.splitlines()[0], 
                        last_wp_report['last_email'], wp_report['datetime']))
        # Save last email datetime if any
        if last_wp_report['last_email']:
            wp_report['last_email']=last_wp_report['last_email']
    
    def format_site(self, wp_site):
        if 'url' not in wp_site :
            log.error("Invalid site %s"%wp_site)
            wp_site={'url':''}
        else:
            # Format sites with scheme indication
            p_url=list(urlparse(wp_site['url']))
            if p_url[0]=="": 
                wp_site['url']='http://'+wp_site['url']
        # Read the wp_site dict and assing default values if needed
        if 'email_to' not in wp_site or wp_site['email_to'] is None: wp_site['email_to']=[]
        if 'false_positive_strings' not in wp_site or wp_site['false_positive_strings'] is None: wp_site['false_positive_strings']=[]
        if 'wpscan_args' not in wp_site or wp_site['wpscan_args'] is None: wp_site['wpscan_args']=[]
        return wp_site

    def scan_site(self, wp_site):
        wp_site=self.format_site(wp_site)
        # Init report variables
        wp_report={
            "site":wp_site['url'],
            "status":None,
            "datetime": datetime.now().strftime('%Y-%m-%dT%H-%M-%S'),
            "last_email":None,
            "errors":[],
            "infos":[],
            "warnings":[],
            "alerts":[],
            "fixed":[],
            "wpscan_output":None # will be deleted
        }

        # Find last site result if any
        last_wp_report=[r for r in self.wp_reports if r['site']==wp_site['url']]
        if len(last_wp_report)>0: 
            last_wp_report=last_wp_report[0]
            # Skip if the daemon mode is enabled and scan already happend in the last configured `daemon_loop_wait`
            if ( self.conf['daemon'] and 
                datetime.strptime(wp_report['datetime'],'%Y-%m-%dT%H-%M-%S') - datetime.strptime(last_wp_report['datetime'],'%Y-%m-%dT%H-%M-%S') < self.conf['daemon_loop_sleep']):
                log.info("Daemon skipping site %s because already scanned in the last %s"%(wp_site['url'] , self.conf['daemon_loop_sleep']))
                self.scanned_sites.append(None)
                return None
        else: last_wp_report=None
        
        # WPScan arguments
        wpscan_arguments=self.conf['wpscan_args']+wp_site['wpscan_args']+['--url', wp_site['url']]

        # Output
        log.info("Scanning site %s"%wp_site['url'] )
        # Launch WPScan -------------------------------------------------------
        (wpscan_exit_code, wp_report["wpscan_output"]) = self.wpscan.wpscan(*wpscan_arguments)

        # Exit code 0: all ok. Exit code 5: Vulnerable. Other exit code are considered as errors
        # Handle scan errors
        if wpscan_exit_code not in [0,5]:
            # Handle API limit
            if "API limit has been reached" in str(wp_report["wpscan_output"]) and self.conf['api_limit_wait']: 
                log.info("API limit has been reached after %s sites, sleeping %s and continuing the scans..."%(len(self.scanned_sites),API_WAIT_SLEEP))
                time.sleep(API_WAIT_SLEEP.total_seconds())
                self.wpscan.update_wpscan()
                return self.scan_site(wp_site)
            # Following redirection
            if "The URL supplied redirects to" in str(wp_report["wpscan_output"]) and self.conf['follow_redirect']: 
                url = re.findall(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+',
                    wp_report["wpscan_output"].split("The URL supplied redirects to")[1] )
                if len(url)==1:
                    wp_site['url']=url[0].strip()
                    log.info("Following redirection to %s"%wp_site['url'])
                    return self.scan_site(wp_site)
                else:
                    err_str="Could not parse the URL to follow in WPScan output after words 'The URL supplied redirects to'"
                    log.error(err_str)
                    wp_report['errors'].append(err_str)

            # Quick return if user cacelled scans ^C or kill
            if wpscan_exit_code in [2]: return None
            log.error("Could not scan site %s"%wp_site['url'])

            # If WPScan error, add the error to the reports
            if wpscan_exit_code in [1,3,4]:  # This types if errors will be written into the Json database file
                err_str="WPScan failed with exit code %s. \nWPScan arguments: %s. \nWPScan output: \n%s"%((wpscan_exit_code, safe_log_wpscan_args(wpscan_arguments), wp_report['wpscan_output']))
                wp_report['errors'].append(err_str)

            # Other errors codes : -9, -2, 127, etc: Just return None right away
            elif not self.conf['fail_fast']: return None 

            # Fail fast
            if self.conf['fail_fast']:
                if not self.interrupting: 
                    log.error("Failure")
                    self.interrupt()
                # Handle the case where the interrupt() method caused WPScan to exit with random codes and not 2
                else: return None 

        # Parse the results if no errors with wpscan -----------------------------
        else:
            # Write wpscan output 
            wpscan_results_file=None
            if self.conf['wpscan_output_folder'] :
                wpscan_results_file=os.path.join(self.conf['wpscan_output_folder'],
                    get_valid_filename('WPScan_results_%s_%s.txt' % (wp_site['url'], wp_report['datetime'])))
                with open(wpscan_results_file, 'w') as wpout:
                    wpout.write(re.sub(r'(\x1b|\[[0-9][0-9]?m)','', str(wp_report['wpscan_output'])))
            
            log.debug("Parsing WPScan output")
            # Call parse_result from wpscan_parser.py ------------------------
            wp_report['infos'], wp_report['warnings'] , wp_report['alerts']  = parse_results(wp_report['wpscan_output'] , 
                self.conf['false_positive_strings']+wp_site['false_positive_strings'] )
            if last_wp_report:
                self.update_report(wp_report, last_wp_report)
            
            # Print WPScan findings ------------------------------------------------------
            for info in wp_report['infos']:
                log.info(oneline("** WPScan INFO %s ** %s" % (wp_site['url'], info )))
            for fix in wp_report['fixed']:
                log.info(oneline("** FIXED %s ** %s" % (wp_site['url'], fix )))
            for warning in wp_report['warnings']:
                log.warning(oneline("** WPScan WARNING %s ** %s" % (wp_site['url'], warning )))
            for alert in wp_report['alerts']:
                log.critical(oneline("** WPScan ALERT %s ** %s" % (wp_site['url'], alert )))

            if wpscan_results_file: log.info("Write wpscan output to file %s"%wpscan_results_file)
        
        # Report status ------------------------------------------------
        if len(wp_report['errors'])>0:wp_report['status']="ERROR"
        elif len(wp_report['warnings'])>0 and len(wp_report['alerts']) == 0: wp_report['status']='WARNING'
        elif len(wp_report['alerts'])>0: wp_report['status']='ALERT'
        elif len(wp_report['fixed'])>0: wp_report['status']='FIXED'
        else: wp_report['status']='INFO'

        # Deleting unwanted informations in report, alerts and fixed items (if any) are always present util next email
        wp_report['warnings']=wp_report['warnings'] if self.conf['send_warnings'] or self.conf['send_infos'] else []
        wp_report['infos']=wp_report['infos'] if self.conf['send_infos'] else []

        # Will print parsed readable Alerts, Warnings, etc as they will appear in email reports
        log.debug("\n"+build_message(wp_report)+"\n")

        # Sending report
        if self.conf['send_email_report']:
            try:
                # Email error report -------------------------------------------------------
                if wp_report['status']=="ERROR":
                    if self.conf['send_errors']:
                        self.send_report(wp_site, wp_report)
                    else:
                        log.info("No WPWatcher ERROR email report have been sent for site %s. If you want to receive error emails, set send_errors=Yes in the config or use --errors."%(wp_site['url']))
                # Or email regular report if conditions ------------------------------------------
                else:
                    if ( self.conf['send_infos'] or 
                        ( wp_report['status']=="WARNING" and self.conf['send_warnings'] ) or 
                        wp_report['status']=='ALERT' or wp_report['status']=='FIXED' ) :

                        if ( not wp_report['last_email'] or ( wp_report['last_email'] and ( 
                            datetime.strptime(wp_report['datetime'],'%Y-%m-%dT%H-%M-%S') - datetime.strptime(wp_report['last_email'],'%Y-%m-%dT%H-%M-%S') > self.conf['resend_emails_after'] 
                            or last_wp_report['status']!=wp_report['status'] ) ) ):
                            # Send the report
                            self.send_report(wp_site, wp_report)
                        else:
                            log.info("Not sending WPWatcher %s email report because already sent in the last %s (at %s) for site %s"%(wp_report['status'], self.conf['resend_emails_after'], wp_report['last_email'], wp_site['url']))
                    else: 
                        # No report notice
                        log.info("No WPWatcher %s email report have been sent for site %s. If you want to receive more emails, send_warnings=Yes, set send_infos=Yes in the config or use --infos."%(wp_report['status'],wp_site['url']))
            
            # Handle send mail error
            except Exception:
                log.error("Unable to send mail report for site " + wp_site['url'] + ". Error: \n"+traceback.format_exc())
                wp_report['errors'].append("Unable to send mail report for site %s"%wp_site['url'])
                if self.conf['fail_fast'] and not self.interrupting: 
                    log.error("Failure")
                    self.interrupt()
        else:
            # No report notice
            log.info("No WPWatcher %s email report have been sent for site %s. To receive emails, setup mail server settings in the config and enable send_email_report or use --send."%(wp_report['status'], wp_site['url']))
        # Save scanned site
        self.scanned_sites.append(wp_site['url'])
        # Discard wpscan_output from report
        del wp_report['wpscan_output']
        # Save report in global instance database and to file when a site has been scanned
        self.update_and_write_wp_reports([wp_report])
        # Print progress
        print_progress_bar(len(self.scanned_sites), len(self.conf['wp_sites'])) 
        return(wp_report)
    
    def interrupt(self, sig=None, frame=None):
        log.error("Interrupting...")

        # Lock for interrupting
        self.interrupting=True
        
        # Cancel all jobs
        for f in self.futures:
            if not f.done(): f.cancel()
        
        # Send ^C to all WPScan
        for p in self.wpscan.processes: 
            p.send_signal(signal.SIGINT)

        # If called inside ThreadPoolExecutor, raise Exeception
        if type(threading.current_thread()) is concurrent.futures.ThreadPoolExecutor:
            raise InterruptedError()
        
        # Wait all scans finished, print results and quit
        else:
            self.executor.shutdown(wait=True)
            self.print_scanned_sites_results()
            log.info("Scans interrupted.")
            exit(-1)

    def print_scanned_sites_results(self):
        if len(self.scanned_sites)>0:
            new_reports=[r for r in self.wp_reports if r['site'] in self.scanned_sites]
            log.info(results_summary(new_reports))

    # Run WPScan on defined websites
    def run_scans_and_notify(self):
        # Check sites are in the config
        if len(self.conf['wp_sites'])==0:
            log.error("No sites configured, please provide wp_sites in config file or use arguments --url URL [URL...] or --urls File path")
            return((-1, self.wp_reports))

        log.info("Starting scans on %s configured sites"%(len(self.conf['wp_sites'])))
        
        new_reports=[]
        
        self.executor=concurrent.futures.ThreadPoolExecutor(max_workers=self.conf['asynch_workers'])
        # Sumbit all scans jobs and start scanning
        for s in self.conf['wp_sites']:
            self.futures.append(self.executor.submit(self.scan_site, s))
        # Loops while scans are running and read results
        for f in self.futures:
            try: new_reports.append(f.result())
            # Handle interruption from inside threads when using --ff
            except (InterruptedError):
                self.executor.shutdown(wait=True)
                self.print_scanned_sites_results()
                log.info("Scans interrupted.")
                return ((-1, self.wp_reports))
        # Print results and finish
        self.print_scanned_sites_results()
        if not any ([r['status']=='ERROR' for r in new_reports if r]):
            log.info("Scans finished successfully.")
            return((0, self.wp_reports))
        else:
            log.info("Scans finished with errors.") 
            return((-1, self.wp_reports))

       


# -*- coding: utf-8 -*-
"""
Wordpress Watcher
Automating WPscan to scan and report vulnerable Wordpress sites

DISCLAIMER - USE AT YOUR OWN RISK.
"""
import logging
# Setup configuration: will be parsed by setup.py -------------------
# Values must be in one line
# Project version.
VERSION='0.5.7b'
# URL that will be displayed in help and other places
GIT_URL="https://github.com/tristanlatr/WPWatcher"
# Authors
AUTHORS="Florian Roth, Tristan Landès"
# Global variables ------------------
# Global log handler
log = logging.getLogger('wpwatcher')



"""
Wordpress Watcher
Automating WPscan to scan and report vulnerable Wordpress sites

DISCLAIMER - USE AT YOUR OWN RISK.
"""
import configparser
import os
import json
from wpwatcher import log, GIT_URL
from wpwatcher.utils import parse_timedelta

# Configuration handling -------------------------------------------------------
class WPWatcherConfig():

    def __init__(self, files=None, string=None):
        self.files=files
        # Init config parser
        self.parser=configparser.ConfigParser()
        # Load default configuration
        self.parser.read_dict({'wpwatcher':self.DEFAULT_CONFIG})

        if string: 
            self.parser.read_string(string)

        if (not self.files or len(self.files)==0) and not string :
            self.files=self.find_config_files()

        if self.files:
            read_files=self.parser.read(self.files)
            if len(read_files) < len(self.files):
                log.error("Could not read config " + str(list(set(self.files)-set(read_files))) + ". Make sure the file exists, the format is OK and you have correct access right.")
                exit(-1)

        # No config file notice
        else :
            log.info("No config file loaded and could not find default config `~/.wpwatcher/wpwatcher.conf`, `~/wpwatcher.conf` or `./wpwatcher.conf`")
    
    def build_config(self):
        config_dict={}
        try:
            # Saving config file in right dict format - no 'wpwatcher' section, just config options
            config_dict = {
                # Configurable witg cli arguments
                'wp_sites' :self.getjson(self.parser,'wp_sites'),
                'send_email_report':self.getbool(self.parser, 'send_email_report'),
                'send_errors':self.getbool(self.parser, 'send_errors'),
                'email_to':self.getjson(self.parser,'email_to'),
                'send_infos':self.getbool(self.parser, 'send_infos'),
                'quiet':self.getbool(self.parser, 'quiet'),
                'verbose':self.getbool(self.parser, 'verbose'),
                'attach_wpscan_output':self.getbool(self.parser, 'attach_wpscan_output'),
                'fail_fast':self.getbool(self.parser, 'fail_fast'),
                'api_limit_wait':self.getbool(self.parser, 'api_limit_wait'),
                'daemon':self.getbool(self.parser, 'daemon'),
                'daemon_loop_sleep':parse_timedelta(self.parser.get('wpwatcher','daemon_loop_sleep')),
                'resend_emails_after':parse_timedelta(self.parser.get('wpwatcher','resend_emails_after')),
                'wp_reports':self.parser.get('wpwatcher','wp_reports'),
                'asynch_workers':self.parser.getint('wpwatcher','asynch_workers'),
                'log_file':self.parser.get('wpwatcher','log_file'),
                'follow_redirect':self.getbool(self.parser, 'follow_redirect'),
                'wpscan_output_folder':self.parser.get('wpwatcher','wpscan_output_folder'),
                'wpscan_args':self.getjson(self.parser,'wpscan_args'),
                # Not configurable with cli arguments
                'send_warnings':self.getbool(self.parser, 'send_warnings'),
                'false_positive_strings' : self.getjson(self.parser,'false_positive_strings'), 
                'email_errors_to':self.getjson(self.parser,'email_errors_to'),
                'wpscan_path':self.parser.get('wpwatcher','wpscan_path'),
                'smtp_server':self.parser.get('wpwatcher','smtp_server'),
                'smtp_auth':self.getbool(self.parser, 'smtp_auth'),
                'smtp_user':self.parser.get('wpwatcher','smtp_user'),
                'smtp_pass':self.parser.get('wpwatcher','smtp_pass'),
                'smtp_ssl':self.getbool(self.parser, 'smtp_ssl'),
                'from_email':self.parser.get('wpwatcher','from_email')
            }
            return ((config_dict, self.files))

        except Exception as err: 
            log.error("Could not read config " + str(self.files) + ". Error: "+str(err))
            raise

    # Configuration template -------------------------
    TEMPLATE_FILE="""[wpwatcher]
# WPWatcher configuration file
# WordPress Watcher is a Python wrapper for WPScan that manages scans on multiple sites and reports by email
# Options configurable with CLI args, see 'wpwatcher --help'
# For more infos check %s

# WPScan configuration
# wpscan_path=/usr/local/rvm/gems/default/wrappers/wpscan
wpscan_args=[   "--format", "cli",
                "--no-banner",
                "--random-user-agent", 
                "--disable-tls-checks" ]

# False positive string matches
# false_positive_strings=["You can get a free API token with 50 daily requests by registering at https://wpvulndb.com/users/sign_up"]

# Sites (--url or --urls)
# wp_sites=   [ {"url":"exemple.com"}, {"url":"exemple2.com"} ]

# Notifications (--send , --em , --infos , --errors , --attach , --resend)
send_email_report=No
email_to=["you@domain"]

# send_infos=Yes
# send_errors=Yes
# send_warnings=No
# attach_wpscan_output=Yes
# resend_emails_after=5d
# email_errors_to=["admins@domain"]

# Email server settings
from_email=WordPressWatcher@domain.com
smtp_server=mailserver.de:587
smtp_auth=Yes
smtp_user=me@domain
smtp_pass=P@assw0rd
smtp_ssl=Yes

# Sleep when API limit reached (--wait)
# api_limit_wait=Yes

# Daemon settings (recommended to use --daemon)
# daemon=No
# daemon_loop_sleep=12h

# Output (-q , -v)
# log_file=/home/user/.wpwatcher/wpwatcher.log
# quiet=Yes
# verbose=Yes
# wpscan_output_folder=/home/user/.wpwatcher/wpscan-results/

# Custom database (--reports)
# wp_reports=/home/user/.wpwatcher/wp_reports.json

# Exit if any errors (--ff)
# fail_fast=Yes 

# Number of asynchronous WPScan executions
# asynch_workers=5

# Follow main redirection when WPScan failed
# follow_redirect=Yes

"""%(GIT_URL)

    # Config default values
    DEFAULT_CONFIG={
        'wp_sites' :'null',
        'false_positive_strings' : 'null',                        
        'wpscan_path':'wpscan',
        'log_file':"",
        'wpscan_args':'''["--no-banner","--random-user-agent"]''',
        'send_email_report':'No',
        'send_errors':'No',
        'email_to':'null',
        'email_errors_to':'null',
        'send_warnings':'Yes',
        'send_infos':'No',
        'attach_wpscan_output':'No',
        'smtp_server':"",
        'smtp_auth':'No',
        'smtp_user':"",
        'smtp_pass':"",
        'smtp_ssl':'No',
        'from_email':"",
        'quiet':'No',
        'verbose':'No',
        'fail_fast':'No',
        'api_limit_wait':'No',
        'daemon':'No',
        'daemon_loop_sleep':'0s',
        'resend_emails_after':'0s',
        'wp_reports':'',
        'asynch_workers':'1',
        'follow_redirect':'No',
        'wpscan_output_folder':''
    }

    @staticmethod
    def getjson(conf, key):
        string_val=conf.get('wpwatcher', key)
        try:
            loaded=json.loads(string_val)
            return loaded if loaded else []
        except Exception as err:
            log.error("Could not read config JSON value for: '%s' and string: '%s'. Error: %s" % (key, conf.get('wpwatcher',key), str(err)))
            raise
    @staticmethod
    def getbool(conf, key):
        try:
            return conf.getboolean('wpwatcher', key)
        except Exception as err:
            log.error("Could not read boolean value in config for: '{}' and string '{}'. Must be Yes/No. Error: {}".format(key, conf.get('wpwatcher',key), str(err)))
            raise
    @staticmethod
    def find_config_files(create=False):
        '''
        Returns the location of existing `wpwatcher.conf` and `wp_reports.json` files at ./wpwatcher.conf and/or ~/wpwatcher.conf or under ~/.wpwatcher/ folder
        '''
        paths=[]
        if create:
            os.makedirs(os.path.join(os.environ['HOME'],'.wpwatcher'), exist_ok=True)
        if 'APPDATA' in os.environ: 
            p=os.path.join(os.environ['APPDATA'],'.wpwatcher/wpwatcher.conf')
            if os.path.isfile(p): paths.append(p)
            elif create: 
                with open(p,'w') as config_file:
                    config_file.write(WPWatcherConfig.TEMPLATE_FILE)
                    log.info("Init new config file from template: %s"%(p))
            p=os.path.join(os.environ['APPDATA'],'wpwatcher.conf')
            if os.path.isfile(p): paths.append(p)
        elif 'HOME' in os.environ: 
            p=os.path.join(os.environ['HOME'],'.wpwatcher/wpwatcher.conf')
            if os.path.isfile(p): paths.append(p)
            elif create: 
                with open(p,'w') as config_file:
                    config_file.write(WPWatcherConfig.TEMPLATE_FILE)
                    log.info("Init new config file: %s"%(p))
            p=os.path.join(os.environ['HOME'],'wpwatcher.conf')
            if os.path.isfile(p): paths.append(p)
        elif 'XDG_CONFIG_HOME' in os.environ: 
            p=os.path.join(os.environ['XDG_CONFIG_HOME'],'.wpwatcher/wpwatcher.conf')
            if os.path.isfile(p): paths.append(p)
            elif create: 
                with open(p,'w') as config_file:
                    config_file.write(WPWatcherConfig.TEMPLATE_FILE)
                    log.info("Init new config file: %s"%(p))
            p=os.path.join(os.environ['XDG_CONFIG_HOME'],'wpwatcher.conf')
            if os.path.isfile(p): paths.append(p)
        if os.path.isfile('./wpwatcher.conf'): 
            paths.append('./wpwatcher.conf')
        return(paths)